#define G_LOG_DOMAIN "boilerplate-plugins-engine"

#include "config.h"

#include "boilerplate-plugins-engine.h"

#include <glib.h>
#include <libpeas/peas.h>

typedef struct _BoilerplatePluginsEnginePrivate
{
    PeasEngine *engine;
    GSettings *settings;
    GHashTable *plugin_settings;
} BoilerplatePluginsEnginePrivate;

struct _BoilerplatePluginsEngine
{
    GObject parent_instance;
};

/**
 * SECTION:boilerplate-plugins-engine
 * @title: BoilerplatePluginsEngine
 * @short_description: Wrapper object for initializing plugins
 * @include: boilerplate-plugins-engine.h
 * @see_also: #PeasEngine
 *
 * #BoilerplatePluginsEngine is a wrapper object for initializing the #PeasEngine.
 **/
G_DEFINE_TYPE_WITH_PRIVATE (BoilerplatePluginsEngine, boilerplate_plugins_engine, G_TYPE_OBJECT)

/* Singleton instance */
BoilerplatePluginsEngine *default_engine = NULL;

/**
 * boilerplate_plugins_engine_add_path:
 * @self: An instance of #BoilerplatePluginsEngine, a sub-object of #PeasEngine
 * @path: The plugin module directory
 * @data_path: (allow-none): The plugin data directory
 *
 * This function appends a search path to the list of paths where #PeasEngine
 * should look for plugins.
 */
void
boilerplate_plugins_engine_add_path (BoilerplatePluginsEngine *self,
                                     const gchar              *path,
                                     const gchar              *data_path)
{
    BoilerplatePluginsEnginePrivate *priv = boilerplate_plugins_engine_get_instance_private (self);

    g_return_if_fail (BOILERPLATE_IS_PLUGINS_ENGINE (self));

    if (path == NULL)
    {
        g_warning ("Plugins path must not be NULL, no path added");
        return;
    }

    peas_engine_add_search_path (PEAS_ENGINE (priv->engine),
	                             path,
	                             data_path);

    g_debug ("Added plugin search path: \"%s\"", path);

    if (data_path != NULL)
        g_debug ("Added plugin data search path: \"%s\"", data_path);
}

/**
 * boilerplate_plugins_engine_add_prepend_path:
 * @self: An instance of #BoilerplatePluginsEngine, a sub-object of #PeasEngine
 * @path: The plugin module directory
 * @data_path: (allow-none): The plugin data directory
 *
 * This function prepends a search path to the list of paths where #PeasEngine
 * should look for plugins.
 */
void
boilerplate_plugins_engine_add_prepend_path (BoilerplatePluginsEngine *self,
                                             const gchar              *path,
                                             const gchar              *data_path)
{
    BoilerplatePluginsEnginePrivate *priv = boilerplate_plugins_engine_get_instance_private (self);

    g_return_if_fail (BOILERPLATE_IS_PLUGINS_ENGINE (self));

    if (path == NULL)
    {
        g_warning ("Plugins path must not be NULL, no path added");
        return;
    }

    peas_engine_prepend_search_path (PEAS_ENGINE (priv->engine),
                                     path,
                                     NULL);

    g_debug ("Added prepended plugin search path: \"%s\"\n", path);

    if (data_path != NULL)
        g_debug ("Added prepended plugin data search path: \"%s\"\n", data_path);
}

/**
 * boilerplate_plugins_engine_get_plugin_settings:
 * @self: An instance of #BoilerplatePluginsEngine, a sub-object of #PeasEngine
 * @module_name: The module name of plugin
 *
 * This functions will return a #GSettings instance for a plugin with
 * @module_name. The settings for a plugin are stored in a #GHashTable for quick
 * access. If this is the first module setting for all plugins then the
 * #GHashTable is created.
 *
 * Individual plugins path are located:
 * |[
 * /tld/domain/name/plugins/module_name/
 * ]|
 *
 * Where `/tld/domain/name` would ideally be your applications resource path and
 * `module_name` is the @module_name paramater.
 *
 * Returns: (transfer full) (GSettings *): A #GSettings instance for a plugin with
 * @module_name
 */
GSettings *
boilerplate_plugins_engine_get_plugin_settings (BoilerplatePluginsEngine *self,
                                                const gchar              *module_name)
{
    BoilerplatePluginsEnginePrivate *priv = boilerplate_plugins_engine_get_instance_private (self);
    GSettings *settings;

    if (G_UNLIKELY (priv->plugin_settings == NULL))
    {
        priv->plugin_settings = g_hash_table_new_full (g_str_hash,
                                                       g_str_equal,
                                                       g_free,
                                                       g_object_unref);
    }

    /* Look for existing settings */
    settings = g_hash_table_lookup (priv->plugin_settings, module_name);

    /* If no settings found, create */
    if (!settings)
    {
        g_autofree gchar *path = NULL;

        path = g_strdup_printf (PACKAGE_RESOURCES "/plugins/%s/", module_name);

        settings = g_settings_new_with_path (PACKAGE_ID ".plugin", path);

        g_hash_table_insert (priv->plugin_settings,
                             g_strdup (module_name),
                             settings);
    }

    return settings;
}

/**
 * boilerplate_plugins_engine_toggle_plugin:
 * @self: An instance of #BoilerplatePluginsEngine, a sub-object of #PeasEngine
 * @key: The key that controls if a plugin is enabled/disabled, this should always
 * be `enabled`, but an override is allowed.
 * @settings: The #GSettings object for the plugin being toggle
 *
 * This function toggles a plugin as enabled/disabled and loads/unloads the
 * plugin in the process. This will be called after your create/get the plugin's
 * settings with boilerplate_plugins_engine_get_plugin_settings().
 *
 * ## Usage
 * |[
 * BoilerplatePluginsEngine *engine = boilerplate_plugins_engine_get_default ();
 *
 * GSettings *example_settings = boilerplate_plugins_engine_get_plugin_settings (engine, "example");
 *
 * boilerplate_plugins_engine_toggle_plugin (engine, "enabled", example_settings);
 * ]|
 */
void
boilerplate_plugins_engine_toggle_plugin (BoilerplatePluginsEngine *self,
                                          const gchar              *key,
                                          GSettings                *settings)
{
    BoilerplatePluginsEnginePrivate *priv = boilerplate_plugins_engine_get_instance_private (self);
    PeasPluginInfo *plugin_info;
    gboolean enabled;

    g_assert (BOILERPLATE_IS_PLUGINS_ENGINE (self));
    g_assert (g_strcmp0 (key, "enabled") == 0);
    g_assert (G_IS_SETTINGS (settings));

    enabled = g_settings_get_boolean (settings, key);

    plugin_info = g_object_get_data (G_OBJECT (settings), "PEAS_PLUGIN_INFO");

    g_assert (plugin_info != NULL);

    if (enabled && !peas_plugin_info_is_loaded (plugin_info))
    {
        peas_engine_load_plugin (priv->engine, plugin_info);
    }
    else if (!enabled && peas_plugin_info_is_loaded (plugin_info))
    {
        peas_engine_unload_plugin (priv->engine, plugin_info);
    }
}

/**
 * boilerplate_plugins_engine_load_plugins:
 * @self: An instance of #BoilerplatePluginsEngine, a sub-object of #PeasEngine
 *
 * This function enables and loads all plugins found in peas_engine_get_plugin_list()
 * from the current default #BoilerplatePluginsEngine instance of @self.
 *
 * ## Usage
 * |[
 * BoilerplatePluginsEngine *engine = boilerplate_plugins_engine_get_default ();
 *
 * boilerplate_plugins_engine_load_plugins (engine);
 * ]|
 */
void
boilerplate_plugins_engine_load_plugins (BoilerplatePluginsEngine *self)
{
    BoilerplatePluginsEnginePrivate *priv = boilerplate_plugins_engine_get_instance_private (self);
    const GList *list;

    list = peas_engine_get_plugin_list (priv->engine);

    for (; list; list = list->next)
    {
        PeasPluginInfo *plugin_info = list->data;
        const gchar *module_name;
        GSettings *settings;

        module_name = peas_plugin_info_get_module_name (plugin_info);
        settings = boilerplate_plugins_engine_get_plugin_settings (self, module_name);

        g_object_set_data (G_OBJECT (settings), "PEAS_PLUGIN_INFO", plugin_info);

        g_signal_connect_object (settings,
                                 "changed::enabled",
                                 G_CALLBACK (boilerplate_plugins_engine_toggle_plugin),
                                 self,
                                 G_CONNECT_SWAPPED);

        if (!g_settings_get_boolean (settings, "enabled"))
            continue;

        g_debug ("Loading plugin \"%s\"",
                 peas_plugin_info_get_module_name (plugin_info));

        peas_engine_load_plugin (priv->engine, plugin_info);
    }
}

/*
 * boilerplate_plugins_engine_dispose:
 * @object: An instance of #BoilerplatePluginsEngine
 *
 * Cleans-up anything that is non-critical for the object.
 */
static void
boilerplate_plugins_engine_dispose (GObject *object)
{
    BoilerplatePluginsEngine *self = BOILERPLATE_PLUGINS_ENGINE (object);
    BoilerplatePluginsEnginePrivate *priv = boilerplate_plugins_engine_get_instance_private (self);

    if (priv->settings != NULL)
    {
        g_object_unref (priv->settings);
    }
    priv->settings = NULL;

    G_OBJECT_CLASS (boilerplate_plugins_engine_parent_class)->dispose (object);
}

/*
 * boilerplate_plugins_engine_finalize:
 * @object: An instance of #BoilerplatePluginsEngine
 *
 * Does any last minute clean-ups when the class ends.
 */
static void
boilerplate_plugins_engine_finalize (GObject *object)
{
    BoilerplatePluginsEngine *self = BOILERPLATE_PLUGINS_ENGINE (object);
    BoilerplatePluginsEnginePrivate *priv = boilerplate_plugins_engine_get_instance_private (self);

    peas_engine_garbage_collect (priv->engine);

    if (priv->engine != NULL)
    {
        g_object_unref (priv->engine);
    }
    priv->engine = NULL;

    G_OBJECT_CLASS (boilerplate_plugins_engine_parent_class)->finalize (object);
}

/*
 * boilerplate_plugins_engine_init:
 * @self: An instance of #BoilerplatePluginsEngine
 *
 * Initializes various aspects of the object.
 * - Creates a #PeasEngine instance
 * - Enables `python3` and `lua5.1` plugins
 * - Adds all search paths
 * - Loads all plugins discovered
 */
static void
boilerplate_plugins_engine_init (BoilerplatePluginsEngine *self)
{
    BoilerplatePluginsEnginePrivate *priv = boilerplate_plugins_engine_get_instance_private (self);
    gchar *path;
    const GList *list;

    priv->engine = peas_engine_get_default ();
    priv->settings = g_settings_new (PACKAGE_ID ".plugins");

    /* Enable 3rd party language loaders */
    peas_engine_enable_loader (PEAS_ENGINE (priv->engine), "python3");
    peas_engine_enable_loader (PEAS_ENGINE (priv->engine), "lua5.1");

    /* Add search directories */
    boilerplate_plugins_engine_add_path (self,
	                                     PACKAGE_PLUGINS_DIRECTORY,
                                         PACKAGE_PLUGINS_DATA_DIRECTORY);

    boilerplate_plugins_engine_add_prepend_path (self,
                                                 "resource://" PACKAGE_RESOURCES "/plugins",
                                                 NULL);

    path = g_build_filename (g_get_home_dir (),
                             ".local",
                             "share",
                             PACKAGE,
                             "plugins",
                             NULL);

    boilerplate_plugins_engine_add_path (self,
                                         path,
                                         NULL);

    g_free (path);

    path = g_build_filename (g_get_user_data_dir (),
                             PACKAGE,
                             "plugins",
                             NULL);

    boilerplate_plugins_engine_add_path (self,
                                         path,
                                         NULL);

    g_free (path);

    /* List plugins that are found */
    list = peas_engine_get_plugin_list (PEAS_ENGINE (priv->engine));

    for (; list; list = list->next)
    {
        PeasPluginInfo *plugin_info = list->data;

        g_debug ("Discovered plugin \"%s\"\n",
                 peas_plugin_info_get_module_name (plugin_info));
    }

    boilerplate_plugins_engine_load_plugins (self);
}

/*
 * boilerplate_plugins_engine_class_init:
 * @klass: An instance of #BoilerplatePluginsEngineClass
 *
 * Initializes the various virtural functions in #BoilerplatePluginsEngine.
 */
static void
boilerplate_plugins_engine_class_init (BoilerplatePluginsEngineClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->dispose = boilerplate_plugins_engine_dispose;
    object_class->finalize = boilerplate_plugins_engine_finalize;
}

/**
 * boilerplate_plugins_engine_new:
 *
 * Creates a new #PeasEngine instance as #BoilerplatePluginsEngine.
 *
 * Returns: #BoilerplatePluginsEngine
 */
BoilerplatePluginsEngine *
boilerplate_plugins_engine_new (void)
{
    return g_object_new (BOILERPLATE_TYPE_PLUGINS_ENGINE,
                         NULL);
}

/**
 * boilerplate_plugins_engine_get_default:
 *
 * Gets the default #PeasEngine, if does not exist create a new instance of
 * #BoilerplatePluginsEngine.
 *
 * Returns: #BoilerplatePluginsEngine
 */
BoilerplatePluginsEngine *
boilerplate_plugins_engine_get_default (void)
{
    if (!default_engine)
    {
        default_engine = boilerplate_plugins_engine_new ();
    }

    return default_engine;
}

